#include <signal.h>
#include <algorithm>
#include <cmath>
#include <msp/fs/utils.h>
#include <msp/gl/blend.h>
#include <msp/gl/framebuffer.h>
#include <msp/gl/matrix.h>
#include <msp/gl/misc.h>
#include <msp/gl/tests.h>
#include <msp/gl/texture2d.h>
#include <msp/input/keys.h>
#include <msp/io/print.h>
#include <msp/stringcodec/codec.h>
#include <msp/stringcodec/utf8.h>
#include <msp/strings/lexicalcast.h>
#include <msp/strings/utils.h>
#include <msp/time/units.h>
#include <msp/time/utils.h>
#include "libr2c2/route.h"
#include "libr2c2/terrain.h"
#include "libr2c2/tracktype.h"
#include "libr2c2/zone.h"
#include "3d/path.h"
#include "designer.h"
#include "elevatetool.h"
#include "extendtool.h"
#include "input.h"
#include "layoutbar.h"
#include "manipulator.h"
#include "measure.h"
#include "movetool.h"
#include "objectproperties.h"
#include "objectselecttool.h"
#include "rotatetool.h"
#include "routebar.h"
#include "selection.h"
#include "slopetool.h"
#include "svgexporter.h"
#include "terraintool.h"
#include "trackbar.h"
#include "zonebar.h"
#include "zoneproperties.h"

using namespace std;
using namespace R2C2;
using namespace Msp;

Designer::Designer(int argc, char **argv):
	window(1280, 960),
	keyboard(window),
	mouse(window),
	ui_res("data/r2c2.res"),
	root(ui_res, &window, &keyboard, &mouse),
	base_object(0),
	cur_route(0),
	cur_zone(0),
	mode(TOOL),
	sel_wrap(selection),
	cur_tool(0),
	keep_status(0)
{
	window.set_title("Railway Designer");
	window.signal_close.connect(sigc::bind(sigc::mem_fun(this, &Designer::exit), 0));

	// Setup catalogue and layout
	catalogue.add_source("data/Märklin/H0");

	cat_layout_3d = new Layout3D(catalogue.get_layout());

	layout = new Layout(catalogue);
	layout->get_clock().set_current_time(12*Time::hour);
	layout_3d = new Layout3D(*layout);
	layout_3d->tick();

	if(argc>1)
	{
		filename = argv[1];
		DataFile::load(*layout, argv[1]);

		if(!layout->get_base().empty())
		{
			base_object = new GL::Object;
			DataFile::load(*base_object, layout->get_base());
		}
	}

	// Setup OpenGL
	cat_view = new View3D(*cat_layout_3d, window.get_width(), window.get_height());

	main_view = new View3D(*layout_3d, window.get_width(), window.get_height());
	GL::Pipeline *pipeline = &main_view->get_pipeline();

	GL::Pipeline::Pass *pass = &pipeline->add_pass("unlit");
	pass->set_depth_test(&GL::DepthTest::lequal());
	pass->set_blend(&GL::Blend::alpha());

	pass = &pipeline->add_pass("blended");
	pass->set_lighting(&layout_3d->get_lighting());
	pass->set_depth_test(&GL::DepthTest::lequal());
	pass->set_blend(&GL::Blend::alpha());

	pass = &pipeline->add_pass("overlay");
	pass->set_blend(&GL::Blend::alpha());

	if(base_object)
		pipeline->add_renderable(*base_object);
	pipeline->add_renderable_for_pass(layout_3d->get_path_scene(), "unlit");
	pipeline->add_renderable_for_pass(layout_3d->get_endpoint_scene(), "unlit");

	// Setup UI
	keyboard.signal_button_press.connect(sigc::bind_return(sigc::mem_fun(this, &Designer::key_press), false));
	mouse.signal_button_press.connect(sigc::bind_return(sigc::mem_fun(this, &Designer::button_press), false));
	root.signal_tooltip.connect(sigc::mem_fun(this, &Designer::tooltip));
	root_layout = new GLtk::Layout;
	root_layout->set_margin(0);
	root.set_layout(root_layout);

	toolbars.push_back(new Layoutbar(*this));
	toolbars.push_back(new Trackbar(*this));
	toolbars.push_back(new Routebar(*this));
	toolbars.push_back(new Zonebar(*this));
	Toolbar *prev = 0;
	for(vector<Toolbar *>::iterator i=toolbars.begin(); i!=toolbars.end(); ++i)
	{
		root.add(**i);
		if(i!=toolbars.begin())
			(*i)->expand(false);
		(*i)->signal_expanded.connect(sigc::hide(sigc::mem_fun(root_layout, &GLtk::Layout::update)));
		if(prev)
		{
			root_layout->add_constraint(**i, GLtk::Layout::RIGHT_OF, *prev, 0);
			root_layout->add_constraint(**i, GLtk::Layout::COPY_HEIGHT, *prev);
		}
		prev = *i;
	}

	GLtk::Panel *statusbar = new GLtk::Panel;
	root.add(*statusbar);
	statusbar->set_focusable(false);
	root_layout->set_gravity(*statusbar, -1, -1);
	root_layout->set_expand(*statusbar, true, false);

	lbl_status = new GLtk::Label;
	statusbar->add(*lbl_status);
	lbl_status->set_geometry(GLtk::Geometry(20, 2, 300, 16));

	pipeline->add_renderable_for_pass(sel_wrap, "unlit");
	overlay = new Overlay3D(ui_res.get_default_font());
	pipeline->add_renderable_for_pass(*overlay, "overlay");

	camera_ctl = new CameraController(*main_view, keyboard, mouse);
	cat_view->get_camera().set_look_direction(GL::Vector3(0, 0.13053, -0.99144));
	cat_view->view_all(true);
	main_view->view_all();

	const set<Object *> &objects = layout->get_all<Object>();
	for(set<Object *>::const_iterator i=objects.begin(); i!=objects.end(); ++i)
		update_object_icon(**i);
}

Designer::~Designer()
{
	for(vector<Toolbar *>::iterator i=toolbars.begin(); i!=toolbars.end(); ++i)
		delete *i;
	delete camera_ctl;
	delete overlay;
	delete main_view;
	delete base_object;
	delete layout_3d;
	delete layout;
	delete cat_view;
	delete cat_layout_3d;
}

int Designer::main()
{
	window.show();

	use_select_tool();

	return Application::main();
}

void Designer::save()
{
	InputDialog *input = new InputDialog(*this, "Save layout", filename);
	input->signal_accept.connect(sigc::mem_fun(layout, &Layout::save));
	root.add(*input);
	root_layout->set_gravity(*input, 0, 0);
}

void Designer::quit()
{
	exit(0);
}

void Designer::new_object()
{
	use_select_tool();

	mode = CATALOGUE;
	lbl_status->set_text("Select new object or press Esc to cancel");
}

void Designer::erase_objects()
{
	use_select_tool();

	set<Object *> objects = selection.get_objects();
	selection.clear();
	for(set<Object *>::iterator i=objects.begin(); i!=objects.end(); ++i)
	{
		overlay->clear(layout_3d->get_3d(**i));
		delete *i;
	}
}

void Designer::object_properties()
{
	use_select_tool();
	if(selection.empty())
		return;

	ObjectProperties *dlg = new ObjectProperties(selection);
	root.add(*dlg);
	root_layout->set_gravity(*dlg, 0, 0);

	dlg->signal_response.connect(sigc::mem_fun(this, &Designer::object_properties_response));
}

void Designer::extend_track()
{
	use_tool_with_selected_objects<ExtendTool>();
}

void Designer::connect_tracks()
{
	use_tool_with_selected_objects<ExtendTool>()->connect();
}

void Designer::flatten_tracks()
{
	use_tool_with_selected_objects<SlopeTool>()->flatten();
}

void Designer::svg_export()
{
	InputDialog *input = new InputDialog(*this, "SVG export", FS::basepart(filename)+".svg");
	input->signal_accept.connect(sigc::mem_fun(this, &Designer::svg_export_accept));
	root.add(*input);
	root_layout->set_gravity(*input, 0, 0);
}

void Designer::edit_route(Route *r)
{
	cur_route = r;
	cur_zone = 0;
	if(cur_route)
		show_route(*cur_route);
	else
		clear_paths();
}

void Designer::rename_route()
{
	use_select_tool();
	if(!cur_route)
		return;

	InputDialog *input = new InputDialog(*this, "Route name", cur_route->get_name());
	input->signal_accept.connect(sigc::mem_fun(this, &Designer::route_name_accept));
	root.add(*input);
	root_layout->set_gravity(*input, 0, 0);
}

void Designer::add_selection_to_route()
{
	if(!cur_route)
		return;

	try
	{
		cur_route->add_tracks(selection.get_objects<Track>());
	}
	catch(const exception &e)
	{
		lbl_status->set_text(e.what());
	}

	show_route(*cur_route);
}

void Designer::edit_zone(Zone *z)
{
	cur_zone = z;
	cur_route = 0;
	if(cur_zone)
		show_zone(*cur_zone);
	else
		clear_paths();
}

void Designer::zone_properties()
{
	if(!cur_zone)
		return;

	ZoneProperties *zone_prop = new ZoneProperties(*cur_zone);
	root.add(*zone_prop);
	const GLtk::Geometry &root_geom = root.get_geometry();
	const GLtk::Geometry &dlg_geom = zone_prop->get_geometry();
	zone_prop->set_position((root_geom.w-dlg_geom.w)/2, (root_geom.h-dlg_geom.h)/2);
}

void Designer::add_selection_to_zone()
{
	if(!cur_zone)
		return;

	try
	{
		cur_zone->add_tracks(selection.get_objects<Track>());
	}
	catch(const exception &e)
	{
		lbl_status->set_text(e.what());
	}

	show_zone(*cur_zone);
}

Vector Designer::map_pointer_to_ground(const Vector &p)
{
	GL::Vector4 vec = main_view->get_camera().unproject(GL::Vector4(p.x, p.y, 0, 0));
	const GL::Vector3 &pos = main_view->get_camera().get_position();

	return pos-GL::Vector3(vec.x, vec.y, vec.z)*(pos.z/vec.z);
}

void Designer::tick()
{
	const Msp::Time::TimeStamp t = Msp::Time::now();
	float dt = (t-last_tick)/Msp::Time::sec;
	last_tick = t;

	if(cur_tool && cur_tool->is_done())
	{
		cur_tool->update_selection(selection);
		use_select_tool();
	}
	keep_status = 0;

	window.tick();
	root.tick();
	camera_ctl->tick(dt);

	render();

	window.swap_buffers();
}

void Designer::key_press(unsigned key)
{
	bool shift = keyboard.get_button_state(Input::KEY_SHIFT_L) || keyboard.get_button_state(Input::KEY_SHIFT_R);

	if(key==Msp::Input::KEY_N && shift)
		use_tool_with_selected_objects<ExtendTool>();
	else if(key==Msp::Input::KEY_N)
		new_object();
	else if(key==Msp::Input::KEY_G)
		use_tool_with_selected_objects<MoveTool>();
	else if(key==Msp::Input::KEY_R)
		use_tool_with_selected_objects<RotateTool>();
	else if(key==Msp::Input::KEY_D)
	{
		const set<Object *> &sel_objs = selection.get_objects();
		list<Object *> new_objs;
		for(set<Object *>::iterator i=sel_objs.begin(); i!=sel_objs.end(); ++i)
		{
			Object *obj = (*i)->clone(layout);
			for(list<Object *>::iterator j=new_objs.begin(); j!=new_objs.end(); ++j)
				obj->link_to(**j);
			new_objs.push_back(obj);
		}

		selection.replace(new_objs.begin(), new_objs.end());

		use_tool_with_selected_objects<MoveTool>();
	}
	else if(key==Msp::Input::KEY_W)
		save();
	else if(key==Msp::Input::KEY_M)
		use_tool<Measure>();
	else if(key==Msp::Input::KEY_Z)
		use_tool_with_selected_objects<ElevateTool>();
	else if(key==Msp::Input::KEY_ESC)
	{
		if(mode==CATALOGUE)
			use_select_tool();
	}
	else if(key==Msp::Input::KEY_X)
		erase_objects();
	else if(key==Msp::Input::KEY_F && shift)
	{
		const set<Track *> &tracks = selection.get_objects<Track>();
		const set<Track *> &ltracks = layout->get_all<Track>();
		for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		{
			(*i)->set_flex(!(*i)->get_flex());
			(*i)->break_links();
			for(set<Track *>::const_iterator j=ltracks.begin(); j!=ltracks.end(); ++j)
				if(*j!=*i)
					(*i)->snap_to(**j, true);

			update_object_icon(**i);
		}
	}
	else if(key==Msp::Input::KEY_F)
		use_tool_with_selected_objects<SlopeTool>()->flatten();
	else if(key==Msp::Input::KEY_E && shift)
		use_tool_with_selected_objects<SlopeTool>()->even_slope(true);
	else if(key==Msp::Input::KEY_E)
		use_tool_with_selected_objects<SlopeTool>()->even_slope();
	else if(key==Msp::Input::KEY_A)
	{
		if(cur_route)
			add_selection_to_route();
		else if(cur_zone)
			add_selection_to_zone();
	}
	else if(key==Msp::Input::KEY_C)
		use_tool_with_selected_objects<ExtendTool>()->connect();
	else if(key==Msp::Input::KEY_V)
		svg_export();
	else if(key==Msp::Input::KEY_P)
		object_properties();
	else if(key==Msp::Input::KEY_TAB)
	{
		Object *obj = selection.get_object();
		if(Terrain *terrain = dynamic_cast<Terrain *>(obj))
			use_tool(new TerrainTool(*this, keyboard, mouse, *terrain));
	}
}

template<typename T>
T *Designer::use_tool_with_selected_objects()
{
	T *tool = new T(*this, keyboard, mouse, selection.get_objects());
	use_tool(tool);
	return tool;
}

template<typename T>
T *Designer::use_tool()
{
	T *tool = new T(*this, keyboard, mouse);
	use_tool(tool);
	return tool;
}

void Designer::use_tool(Tool *tool)
{
	if(cur_tool)
	{
		delete cur_tool;
		cur_tool = 0;
	}

	cur_tool = tool;
	cur_tool->signal_status.connect(sigc::mem_fun(this, &Designer::tool_status));
	if(keep_status<2)
		tool_status(tool->get_status());
	mode = TOOL;
	keep_status = 1;
}

void Designer::use_select_tool()
{
	use_tool(new ObjectSelectTool(*this, keyboard, mouse, selection));
}

void Designer::button_press(unsigned btn)
{
	Vector pointer(mouse.get_axis_value(0), mouse.get_axis_value(1), 0);
	Vector ground = map_pointer_to_ground(pointer);

	if(mode==CATALOGUE)
	{
		Object *obj = 0;
		if(btn==1)
			obj = pick_object(pointer);

		if(obj)
		{
			obj = obj->clone(layout);
			obj->set_position(ground);

			selection.clear();
			selection.add_object(obj);
		}

		if(obj || btn==3)
			use_select_tool();
	}
}

void Designer::render()
{
	GL::Framebuffer::system().clear(GL::COLOR_BUFFER_BIT|GL::DEPTH_BUFFER_BIT);

	View3D *view = (mode==CATALOGUE ? cat_view : main_view);
	view->render();

	root.render();
}

Object *Designer::pick_object(const Vector &pointer)
{
	View3D &view = *(mode==CATALOGUE ? cat_view : main_view);
	const GL::Vector3 &cpos = view.get_camera().get_position();
	GL::Vector4 cray = view.get_camera().unproject(GL::Vector4(pointer.x, pointer.y, 0, 0));

	return view.get_layout().get_layout().pick<Object>(Ray(cpos, cray.slice<3>(0)));
}

void Designer::update_object_icon(Object &obj)
{
	Object3D &obj3d = layout_3d->get_3d(obj);
	overlay->clear(obj3d);

	if(Track *track = dynamic_cast<Track *>(&obj))
	{
		if(track->get_flex())
			overlay->add_graphic(obj3d, "flex");

		if(unsigned saddr = track->get_sensor_address())
		{
			overlay->add_graphic(obj3d, "trackcircuit");
			overlay->set_label(obj3d, lexical_cast<string>(saddr));
		}
		else if(track->get_type().is_turnout())
		{
			unsigned taddr = track->get_turnout_address();
			if(taddr<0x800)
			{
				overlay->add_graphic(obj3d, "turnout");
				overlay->set_label(obj3d, lexical_cast<string>(taddr));
			}
		}
	}
	else if(Signal *signal = dynamic_cast<Signal *>(&obj))
	{
		overlay->add_graphic(obj3d, "signal");
		overlay->set_label(obj3d, lexical_cast<string>(signal->get_address()));
	}
	else if(BeamGate *gate = dynamic_cast<BeamGate *>(&obj))
	{
		overlay->add_graphic(obj3d, "beamgate");
		overlay->set_label(obj3d, lexical_cast<string>(gate->get_address()));
	}
}

void Designer::tool_status(const string &status)
{
	lbl_status->set_text(status);
	if(keep_status==1)
		keep_status = 2;
}

void Designer::object_properties_response(int)
{
	const set<Object *> &objects = selection.get_objects();
	for(set<Object *>::const_iterator i=objects.begin(); i!=objects.end(); ++i)
		update_object_icon(**i);
}

void Designer::route_name_accept(const string &text)
{
	if(cur_route)
		cur_route->set_name(text);
}

void Designer::svg_export_accept(const string &text)
{
	SvgExporter svg_exp(*layout);
	svg_exp.save(text);
}

string Designer::tooltip(int x, int y)
{
	const GLtk::Geometry &rgeom = root.get_geometry();
	if(Object *obj = pick_object(Vector(x*2.0f/rgeom.w-1.0f, y*2.0f/rgeom.h-1.0f, 0)))
	{
		const ObjectType &otype = obj->get_type();
		string info = format("%d %s", otype.get_article_number(), otype.get_name());
		if(Track *track = dynamic_cast<Track *>(obj))
		{
			if(mode!=CATALOGUE && abs(track->get_tilt()).radians()>1e-4)
				info += format(" (slope %.1f%%)", abs(tan(track->get_tilt())*100));
			if(track->get_type().is_turnout())
				info += format(" (turnout %d)", track->get_turnout_address());
			else if(track->get_sensor_address())
				info += format(" (sensor %d)", track->get_sensor_address());
		}
		if(mode==CATALOGUE)
		{
			const string &descr = otype.get_description();
			if(!descr.empty())
			{
				info += '\n';
				info += otype.get_description();
			}
		}
		return info;
	}

	return string();
}

void Designer::clear_paths()
{
	for(list<Path3D *>::iterator i=highlight_paths.begin(); i!=highlight_paths.end(); ++i)
		delete *i;
	highlight_paths.clear();
}

void Designer::show_route(const Route &route)
{
	clear_paths();

	const set<Track *> &rtracks = route.get_tracks();
	for(set<Track *>::iterator i=rtracks.begin(); i!=rtracks.end(); ++i)
	{
		Track3D &t3d = layout_3d->get_3d(**i);
		Path3D *path = new Path3D(t3d);
		path->set_color(GL::Color(0.5, 0.8, 1.0));
		if((*i)->get_type().is_turnout())
			path->set_path(route.get_turnout((*i)->get_turnout_address()));
		highlight_paths.push_back(path);
	}
}

void Designer::show_zone(const Zone &zone)
{
	clear_paths();

	const Zone::TrackSet &ztracks = zone.get_tracks();
	for(Zone::TrackSet::const_iterator i=ztracks.begin(); i!=ztracks.end(); ++i)
	{
		Track3D &t3d = layout_3d->get_3d(**i);
		Path3D *path = new Path3D(t3d);
		path->set_color(GL::Color(0.8, 1.0, 0.5));
		highlight_paths.push_back(path);
	}
}
