#include <msp/core/maputils.h>
#include <msp/datafile/writer.h>
#include <msp/fs/redirectedpath.h>
#include <msp/fs/stat.h>
#include <msp/io/print.h>
#include <msp/time/utils.h>
#include "arducontrol.h"
#include "tracktype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

ArduControl::ProtocolInfo ArduControl::protocol_info[2] =
{
	{ 79, 14, 4 },       // MM
	{ 0x3FFF, 126, 15 }  // MFX
};

Driver::TelemetryInfo ArduControl::telemetry_info[4] =
{
	{ "voltage", "Voltage", "V", 1 },
	{ "current", "Current", "A", 2 },
	{ "cmd-queue-depth", "Cmd queue", "", 0 },
	{ "s88-latency", "S88 latency", "ms", 0 }
};

ArduControl::ArduControl(const Options &opts):
	serial(opts.get<string>(string(), "ttyUSB0")),
	debug(opts.get<unsigned>("debug")),
	state_file("arducontrol.state"),
	power(false),
	halted(false),
	active_accessory(0),
	command_timeout(200*Time::msec),
	s88(*this),
	mfx_search(*this),
	thread(*this)
{
	if(FS::exists(state_file))
		DataFile::load(*this, state_file.str());

	unsigned max_address = 0;
	for(MfxInfoArray::const_iterator i=mfx_info.begin(); i!=mfx_info.end(); ++i)
		max_address = max(max_address, i->address);
	mfx_search.set_next_address(max_address+1);

	PendingCommand cmd;
	cmd.command[0] = READ_POWER_STATE;
	cmd.length = 1;
	command_queue.push(cmd);

	cmd.command[0] = MFX_SET_STATION_ID;
	cmd.command[1] = 'R';
	cmd.command[2] = '2';
	cmd.command[3] = 'C';
	cmd.command[4] = '2';
	cmd.length = 5;
	command_queue.push(cmd);
}

ArduControl::~ArduControl()
{
	thread.exit();
}

void ArduControl::set_power(bool p)
{
	if(power.set(p))
	{
		PendingCommand cmd(POWER);
		cmd.tag.serial = power.serial;
		cmd.command[0] = (p ? POWER_ON : POWER_OFF);
		cmd.length = 1;
		command_queue.push(cmd);
	}
}

void ArduControl::halt(bool h)
{
	if(h==halted)
		return;

	halted = h;
	if(halted)
	{
		for(LocomotiveMap::const_iterator i=locomotives.begin(); i!=locomotives.end(); ++i)
			set_loco_speed(i->first, 0);
	}

	signal_halt.emit(halted);
}

const char *ArduControl::enumerate_protocols(unsigned i) const
{
	if(i==0)
		return "MM";
	else if(i==1)
		return "MFX";
	else
		return 0;
}

ArduControl::Protocol ArduControl::map_protocol(const string &proto_name)
{
	if(proto_name=="MM")
		return MM;
	else if(proto_name=="MFX")
		return MFX;
	else
		throw invalid_argument("ArduControl::map_protocol");
}

unsigned ArduControl::get_protocol_speed_steps(const string &proto_name) const
{
	return protocol_info[map_protocol(proto_name)].max_speed;
}

const Driver::DetectedLocomotive *ArduControl::enumerate_detected_locos(unsigned i) const
{
	if(i>=mfx_info.size())
		return 0;

	return &mfx_info[i];
}

unsigned ArduControl::add_loco(unsigned addr, const string &proto_name, const VehicleType &)
{
	if(!addr)
		throw invalid_argument("ArduControl::add_loco");

	Protocol proto = map_protocol(proto_name);
	if(addr>protocol_info[proto].max_address)
		throw invalid_argument("ArduControl::add_loco");

	Locomotive loco(proto, addr);
	insert_unique(locomotives, loco.id, loco);

	return loco.id;
}

ArduControl::MfxInfoArray::iterator ArduControl::add_mfx_info(const MfxInfo &info)
{
	MfxInfoArray::iterator i;
	for(i=mfx_info.begin(); (i!=mfx_info.end() && i->id!=info.id); ++i) ;
	if(i==mfx_info.end())
	{
		mfx_info.push_back(info);
		i = --mfx_info.end();
	}
	else
		*i = info;
	return i;
}

void ArduControl::remove_loco(unsigned id)
{
	Locomotive &loco = get_item(locomotives, id);
	refresh.remove_loco(loco);
	locomotives.erase(id);
}

void ArduControl::set_loco_speed(unsigned id, unsigned speed)
{
	Locomotive &loco = get_item(locomotives, id);
	if(speed>protocol_info[loco.proto].max_speed)
		throw invalid_argument("ArduControl::set_loco_speed");

	if(speed && halted)
		return;

	if(loco.speed.set(speed))
	{
		PendingCommand cmd(loco, Locomotive::SPEED);
		command_queue.push(cmd);

		refresh.add_loco(loco);
	}
}

void ArduControl::set_loco_reverse(unsigned id, bool rev)
{
	Locomotive &loco = get_item(locomotives, id);
	if(loco.reverse.set(rev))
	{
		PendingCommand cmd(loco, Locomotive::REVERSE);
		command_queue.push(cmd);

		refresh.add_loco(loco);
	}
}

void ArduControl::set_loco_function(unsigned id, unsigned func, bool state)
{
	Locomotive &loco = get_item(locomotives, id);
	if(func>protocol_info[loco.proto].max_func)
		throw invalid_argument("ArduControl::set_loco_function");

	unsigned mask = 1<<func;
	if(loco.funcs.set((loco.funcs&~mask)|(mask*state)))
	{
		if(func>0 || loco.proto!=MM)
		{
			PendingCommand cmd(loco, Locomotive::FUNCTIONS, func);
			command_queue.push(cmd);
		}

		refresh.add_loco(loco);
	}
}

unsigned ArduControl::add_turnout(unsigned addr, const TrackType &type)
{
	if(!addr || !type.is_turnout())
		throw invalid_argument("ArduControl::add_turnout");

	return add_accessory(Accessory::TURNOUT, addr, type.get_state_bits(), type.get_paths());
}

void ArduControl::remove_turnout(unsigned addr)
{
	remove_accessory(Accessory::TURNOUT, addr);
}

void ArduControl::set_turnout(unsigned addr, unsigned state)
{
	set_accessory(Accessory::TURNOUT, addr, state);
}

unsigned ArduControl::get_turnout(unsigned addr) const
{
	return get_accessory(Accessory::TURNOUT, addr);
}

unsigned ArduControl::add_signal(unsigned addr, const SignalType &)
{
	return add_accessory(Accessory::SIGNAL, addr, 1, 3);
}

void ArduControl::remove_signal(unsigned addr)
{
	remove_accessory(Accessory::SIGNAL, addr);
}

void ArduControl::set_signal(unsigned addr, unsigned state)
{
	set_accessory(Accessory::SIGNAL, addr, state);
}

unsigned ArduControl::get_signal(unsigned addr) const
{
	return get_accessory(Accessory::SIGNAL, addr);
}

unsigned ArduControl::add_accessory(Accessory::Kind kind, unsigned addr, unsigned bits, unsigned states)
{
	AccessoryMap::iterator i = accessories.lower_bound(addr);
	AccessoryMap::iterator j = accessories.upper_bound(addr+bits-1);
	if(i!=j)
		throw key_error(addr);
	if(i!=accessories.begin())
	{
		--i;
		if(i->first+i->second.bits>addr)
			throw key_error(addr);
	}

	insert_unique(accessories, addr, Accessory(kind, addr, bits, states));
	return addr;
}

void ArduControl::remove_accessory(Accessory::Kind kind, unsigned addr)
{
	Accessory &acc = get_item(accessories, addr);
	if(acc.kind!=kind)
		throw key_error(addr);
	accessories.erase(addr);
}

void ArduControl::set_accessory(Accessory::Kind kind, unsigned addr, unsigned state)
{
	Accessory &acc = get_item(accessories, addr);
	if(acc.kind!=kind)
		throw key_error(addr);

	if(state!=acc.target || acc.uncertain)
	{
		acc.target = state;
		accessory_queue.push_back(&acc);
	}
}

unsigned ArduControl::get_accessory(Accessory::Kind kind, unsigned addr) const
{
	const Accessory &acc = get_item(accessories, addr);
	if(acc.kind!=kind)
		throw key_error(addr);
	return acc.state;
}

void ArduControl::activate_accessory_by_mask(Accessory &acc, unsigned mask)
{
	unsigned bit = mask&~(mask-1);
	for(active_index=0; (bit>>active_index)>1; ++active_index) ;
	acc.state.set((acc.state&~bit)|(acc.target&bit));
	if(debug>=1)
		IO::print("Setting accessory %d bit %d, state=%d\n", acc.address, active_index, acc.state.pending);
	PendingCommand cmd(acc, Accessory::ACTIVATE, active_index);
	command_queue.push(cmd);
	active_accessory = &acc;

	monitor.reset_peak();
}

unsigned ArduControl::add_sensor(unsigned addr)
{
	if(!addr)
		throw invalid_argument("ArduControl::add_sensor");

	insert_unique(sensors, addr, Sensor(addr));
	s88.grow_n_octets((addr+7)/8);

	return addr;
}

void ArduControl::remove_sensor(unsigned addr)
{
	remove_existing(sensors, addr);
	// TODO update s88.n_octets
}

bool ArduControl::get_sensor(unsigned addr) const
{
	return get_item(sensors, addr).state;
}

const Driver::TelemetryInfo *ArduControl::enumerate_telemetry(unsigned i) const
{
	if(i<4)
		return telemetry_info+i;
	else
		return 0;
}

float ArduControl::get_telemetry_value(const string &name) const
{
	if(name==telemetry_info[0].name)
		return monitor.get_voltage();
	else if(name==telemetry_info[1].name)
		return monitor.get_current();
	else if(name==telemetry_info[2].name)
		return command_queue.size();
	else if(name==telemetry_info[3].name)
		return s88.get_latency()/Time::msec;
	else
		throw key_error(name);
}

void ArduControl::tick()
{
	Tag tag;
	while(completed_commands.pop(tag))
	{
		if(tag.type==Tag::GENERAL)
		{
			if(tag.command==POWER)
			{
				if(power.commit(tag.serial))
					signal_power.emit(power.current);
			}
			else if(tag.command==NEW_LOCO)
			{
				MfxInfo info;
				if(mfx_search.pop_info(info))
				{
					MfxInfoArray::iterator i = add_mfx_info(info);
					save_state();
					signal_locomotive_detected.emit(*i);
				}
			}
		}
		else if(tag.type==Tag::LOCOMOTIVE)
		{
			LocomotiveMap::iterator i = locomotives.find(tag.id);
			if(i==locomotives.end())
				continue;

			Locomotive &loco = i->second;
			if(tag.command==Locomotive::SPEED)
			{
				if(loco.speed.commit(tag.serial))
					signal_loco_speed.emit(loco.id, loco.speed, loco.reverse);
			}
			else if(tag.command==Locomotive::REVERSE)
			{
				if(loco.reverse.commit(tag.serial))
					signal_loco_speed.emit(loco.id, loco.speed, loco.reverse);
			}
			else if(tag.command==Locomotive::FUNCTIONS)
			{
				unsigned old = loco.funcs;
				if(loco.funcs.commit(tag.serial))
				{
					unsigned changed = old^loco.funcs;
					for(unsigned j=0; changed>>j; ++j)
						if((changed>>j)&1)
							signal_loco_function.emit(loco.id, j, (loco.funcs>>j)&1);
				}
			}
		}
		else if(tag.type==Tag::ACCESSORY)
		{
			AccessoryMap::iterator i = accessories.find(tag.id);
			if(i==accessories.end())
				continue;

			Accessory &acc = i->second;
			if(tag.command==Accessory::ACTIVATE)
				off_timeout = Time::now()+acc.active_time;
			else if(tag.command==Accessory::DEACTIVATE)
			{
				if(acc.state.commit(tag.serial))
				{
					if(&acc==active_accessory)
						active_accessory = 0;
				}
			}
		}
		else if(tag.type==Tag::SENSOR)
		{
			SensorMap::iterator i = sensors.find(tag.id);
			if(i==sensors.end())
				continue;

			Sensor &sensor = i->second;
			if(tag.command==Sensor::STATE)
			{
				if(sensor.state.commit(tag.serial))
					signal_sensor.emit(sensor.address, sensor.state);
			}
		}
	}

	while(power && !active_accessory && !accessory_queue.empty())
	{
		Accessory &acc = *accessory_queue.front();

		if(acc.uncertain)
		{
			unsigned zeroes = acc.uncertain&~acc.target;
			if(zeroes)
				activate_accessory_by_mask(acc, zeroes);
			else
				activate_accessory_by_mask(acc, acc.uncertain);
		}
		else if(acc.state!=acc.target)
		{
			unsigned changes = acc.state^acc.target;
			if(!(changes&((1<<acc.bits)-1)))
			{
				// All remaining changes are in non-physical bits
				acc.state.set(acc.state^changes);
				acc.state.commit(acc.state.serial);
			}
			else
			{
				unsigned toggle_bit = 0;
				for(unsigned bit=1; (!toggle_bit && bit<=changes); bit<<=1)
					if((changes&bit) && (acc.valid_states&(1<<(acc.state^bit))))
						toggle_bit = bit;

				activate_accessory_by_mask(acc, toggle_bit);
			}
		}
		else
		{
			accessory_queue.pop_front();

			if(acc.state==acc.target)
			{
				if(acc.kind==Accessory::TURNOUT)
					signal_turnout.emit(acc.address, acc.state);
				else if(acc.kind==Accessory::SIGNAL)
					signal_signal.emit(acc.address, acc.state);
			}
		}
	}

	if(active_accessory && off_timeout)
	{
		bool success = (monitor.get_peak()>0.35f && monitor.get_current()<monitor.get_peak()-0.2f);
		Time::TimeStamp t = Time::now();
		if(t>off_timeout || success)
		{
			Accessory &acc = *active_accessory;

			unsigned bit = 1<<active_index;

			// Assume success if we were uncertain of the physical setting
			if(acc.uncertain&bit)
				acc.uncertain &= ~bit;
			else if(acc.kind==Accessory::TURNOUT && !success)
			{
				if(debug>=1)
					IO::print("Peak current only %.2f A\n", monitor.get_peak());
				signal_turnout_failed.emit(acc.address);
				acc.state.rollback();
				if(acc.valid_states&(1<<(acc.target^bit)))
					acc.target ^= bit;
				else
					acc.target = acc.state;
			}

			off_timeout = Time::TimeStamp();
			PendingCommand cmd(acc, Accessory::DEACTIVATE, active_index);
			command_queue.push(cmd);
		}
	}
}

void ArduControl::flush()
{
	while(!command_queue.empty() || (power && !accessory_queue.empty()))
		tick();
}

void ArduControl::save_state() const
{
	FS::RedirectedPath tmp_file(state_file);
	IO::BufferedFile out(tmp_file.str(), IO::M_WRITE);
	DataFile::Writer writer(out);

	writer.write((DataFile::Statement("mfx_announce_serial"), mfx_announce.get_serial()));
	for(MfxInfoArray::const_iterator i=mfx_info.begin(); i!=mfx_info.end(); ++i)
	{
		DataFile::Statement st("mfx_locomotive");
		st.append(i->id);
		st.sub.push_back((DataFile::Statement("address"), i->address));
		st.sub.push_back((DataFile::Statement("name"), i->name));
		writer.write(st);
	}
}


ArduControl::Tag::Tag():
	type(NONE),
	command(0),
	serial(0),
	id(0)
{ }


ArduControl::Locomotive::Locomotive(Protocol p, unsigned a):
	id((p<<16)|a),
	proto(p),
	address(a),
	speed(0),
	reverse(false),
	funcs(0),
	last_change_age(0)
{ }

unsigned ArduControl::Locomotive::create_speed_dir_command(char *buffer) const
{
	if(proto==MM)
	{
		buffer[0] = MOTOROLA_SPEED_DIRECTION;
		buffer[1] = address;
		buffer[2] = funcs.pending&1;
		buffer[3] = speed.pending+reverse.pending*0x80;
		return 4;
	}
	else if(proto==MFX)
	{
		buffer[0] = MFX_SPEED;
		buffer[1] = address>>8;
		buffer[2] = address;
		buffer[3] = speed.pending+reverse.pending*0x80;
		return 4;
	}
	else
		return 0;
}

unsigned ArduControl::Locomotive::create_speed_func_command(unsigned f, char *buffer) const
{
	if(proto==MM)
	{
		if(f<1 || f>4)
			throw invalid_argument("Locomotive::create_speed_func_command");

		buffer[0] = MOTOROLA_SPEED_FUNCTION;
		buffer[1] = address;
		buffer[2] = (f<<4)|(((funcs.pending>>f)&1)<<1)|(funcs.pending&1);
		buffer[3] = speed.pending;
		return 4;
	}
	else if(proto==MFX)
	{
		bool f16 = (funcs.pending>0xFF);
		buffer[0] = (f16 ? MFX_SPEED_FUNCS16 : MFX_SPEED_FUNCS8);
		buffer[1] = address>>8;
		buffer[2] = address;
		buffer[3] = speed.pending+reverse.pending*0x80;
		if(f16)
		{
			buffer[4] = funcs.pending>>8;
			buffer[5] = funcs.pending;
			return 6;
		}
		else
		{
			buffer[4] = funcs.pending;
			return 5;
		}
	}
	else
		return 0;
}


ArduControl::Accessory::Accessory(Kind k, unsigned a, unsigned b, unsigned s):
	kind(k),
	address(a),
	bits(b),
	valid_states(s),
	state(0),
	uncertain((1<<bits)-1),
	target(0),
	active_time((bits*700)*Time::msec)
{ }

unsigned ArduControl::Accessory::create_state_command(unsigned b, bool c, char *buffer) const
{
	if(b>=bits)
		throw invalid_argument("Accessory::create_state_command");

	unsigned a = (address+b+3)*2;
	if(!((state.pending>>b)&1))
		++a;
	buffer[0] = MOTOROLA_SOLENOID;
	buffer[1] = a>>3;
	buffer[2] = ((a&7)<<4)|c;
	return 3;
}


ArduControl::Sensor::Sensor(unsigned a):
	address(a),
	state(false)
{ }


ArduControl::PendingCommand::PendingCommand():
	length(0),
	repeat_count(1)
{ }

ArduControl::PendingCommand::PendingCommand(GeneralCommand cmd):
	length(0),
	repeat_count(1)
{
	tag.type = Tag::GENERAL;
	tag.command = cmd;
}

ArduControl::PendingCommand::PendingCommand(Locomotive &loco, Locomotive::Command cmd, unsigned index):
	repeat_count(8)
{
	tag.type = Tag::LOCOMOTIVE;
	tag.command = cmd;
	tag.id = loco.id;
	if(cmd==Locomotive::SPEED)
	{
		tag.serial = loco.speed.serial;
		length = loco.create_speed_dir_command(command);
	}
	else if(cmd==Locomotive::REVERSE)
	{
		tag.serial = loco.reverse.serial;
		length = loco.create_speed_dir_command(command);
	}
	else if(cmd==Locomotive::FUNCTIONS)
	{
		tag.serial = loco.funcs.serial;
		length = loco.create_speed_func_command(index, command);
	}
	else
		throw invalid_argument("PendingCommand");
}

ArduControl::PendingCommand::PendingCommand(Accessory &acc, Accessory::Command cmd, unsigned index):
	repeat_count(1)
{
	tag.type = Tag::ACCESSORY;
	tag.command = cmd;
	tag.id = acc.address;
	if(cmd==Accessory::ACTIVATE || cmd==Accessory::DEACTIVATE)
	{
		tag.serial = acc.state.serial;
		length = acc.create_state_command(index, (cmd==Accessory::ACTIVATE), command);
	}
	else
		throw invalid_argument("PendingCommand");
}


template<typename T>
void ArduControl::Queue<T>::push(const T &item)
{
	MutexLock lock(mutex);
	items.push_back(item);
}

template<typename T>
bool ArduControl::Queue<T>::pop(T &item)
{
	MutexLock lock(mutex);
	if(items.empty())
		return false;

	item = items.front();
	items.pop_front();
	return true;
}

template<typename T>
unsigned ArduControl::Queue<T>::size() const
{
	return items.size();
}

template<typename T>
bool ArduControl::Queue<T>::empty() const
{
	return items.empty();
}


bool ArduControl::CommandQueueTask::get_work(PendingCommand &cmd)
{
	return queue.pop(cmd);
}

void ArduControl::CommandQueueTask::push(const PendingCommand &cmd)
{
	queue.push(cmd);
}


ArduControl::Task::Task(const string &n, unsigned p):
	name(n),
	priority(p)
{ }

void ArduControl::Task::sleep(const Time::TimeDelta &dt)
{
	sleep_timeout = Time::now()+dt;
}


ArduControl::CommandQueueTask::CommandQueueTask():
	Task("CommandQueue")
{ }


ArduControl::RefreshTask::RefreshTask():
	Task("Refresh", 2),
	next(cycle.end()),
	round(0),
	loco(0),
	phase(0)
{ }

bool ArduControl::RefreshTask::get_work(PendingCommand &cmd)
{
	if(loco && loco->proto==MM && phase==0)
	{
		cmd.length = loco->create_speed_func_command(round%4+1, cmd.command);
		cmd.repeat_count = 2;
		++phase;
		return true;
	}

	loco = get_next_loco();
	if(!loco)
		return false;

	phase = 0;
	if(loco->proto==MM)
	{
		cmd.length = loco->create_speed_dir_command(cmd.command);
		cmd.repeat_count = 2;
	}
	else if(loco->proto==MFX)
		cmd.length = loco->create_speed_func_command(0, cmd.command);
	else
		return false;

	return true;
}

void ArduControl::RefreshTask::add_loco(Locomotive &l)
{
	MutexLock lock(mutex);
	cycle.push_back(&l);
	if(cycle.size()>15)
	{
		LocomotivePtrList::iterator oldest = cycle.begin();
		for(LocomotivePtrList::iterator i=cycle.begin(); ++i!=cycle.end(); )
			if((*i)->last_change_age>(*oldest)->last_change_age)
				oldest = i;
		if(oldest==next)
			advance();
		cycle.erase(oldest);
	}
	if(next==cycle.end())
		next = cycle.begin();
}

void ArduControl::RefreshTask::remove_loco(Locomotive &l)
{
	MutexLock lock(mutex);
	for(LocomotivePtrList::iterator i=cycle.begin(); i!=cycle.end(); ++i)
		if(*i==&l)
		{
			if(i==next)
			{
				if(cycle.size()>1)
					advance();
				else
					next = cycle.end();
			}
			cycle.erase(i);
			return;
		}
}

ArduControl::Locomotive *ArduControl::RefreshTask::get_next_loco()
{
	MutexLock lock(mutex);
	if(cycle.empty())
		return 0;

	Locomotive *l = *next;
	advance();
	return l;
}

void ArduControl::RefreshTask::advance()
{
	++next;
	if(next==cycle.end())
	{
		next = cycle.begin();
		++round;
	}
}


ArduControl::S88Task::S88Task(ArduControl &c):
	Task("S88"),
	control(c),
	n_octets(0),
	octets_remaining(0)
{ }

bool ArduControl::S88Task::get_work(PendingCommand &cmd)
{
	if(octets_remaining || !n_octets)
		return false;

	Time::TimeStamp t = Time::now();
	if(last_poll)
		latency = t-last_poll;
	last_poll = t;

	octets_remaining = n_octets;
	cmd.command[0] = S88_READ;
	cmd.command[1] = octets_remaining;
	cmd.length = 2;

	sleep(100*Time::msec);

	return true;
}

void ArduControl::S88Task::process_reply(const char *reply, unsigned length)
{
	unsigned char type = reply[0];
	if(type==S88_DATA && length>2)
	{
		unsigned offset = static_cast<unsigned char>(reply[1]);
		unsigned count = length-2;

		SensorMap::iterator begin = control.sensors.lower_bound(offset*8+1);
		SensorMap::iterator end = control.sensors.upper_bound((offset+count)*8);
		for(SensorMap::iterator i=begin; i!=end; ++i)
		{
			unsigned bit_index = i->first-1-offset*8;
			bool state = (reply[2+bit_index/8]>>(7-bit_index%8))&1;
			i->second.state.set(state);

			Tag tag;
			tag.type = Tag::SENSOR;
			tag.command = Sensor::STATE;
			tag.serial = i->second.state.serial;
			tag.id = i->first;
			control.completed_commands.push(tag);
		}

		if(count>octets_remaining)
			octets_remaining = 0;
		else
			octets_remaining -= count;
	}
}

void ArduControl::S88Task::set_n_octets(unsigned n)
{
	n_octets = n;
}

void ArduControl::S88Task::grow_n_octets(unsigned n)
{
	if(n>n_octets)
		n_octets = n;
}


ArduControl::MfxAnnounceTask::MfxAnnounceTask():
	Task("MfxAnnounce", 1),
	serial(0)
{ }

bool ArduControl::MfxAnnounceTask::get_work(PendingCommand &cmd)
{
	cmd.command[0] = MFX_ANNOUNCE;
	cmd.command[1] = serial>>8;
	cmd.command[2] = serial;
	cmd.length = 3;

	sleep(400*Time::msec);

	return true;
}

void ArduControl::MfxAnnounceTask::set_serial(unsigned s)
{
	serial = s;
}


ArduControl::MfxSearchTask::MfxSearchTask(ArduControl &c):
	Task("MfxSearch", 1),
	control(c),
	next_address(1),
	size(0),
	bits(0),
	misses(0)
{ }

bool ArduControl::MfxSearchTask::get_work(PendingCommand &cmd)
{
	if(size>32)
	{
		if(control.debug>=1)
			IO::print("Assigning MFX address %d to decoder %08X\n", next_address, bits);

		MfxInfo info;
		info.protocol = "MFX";
		info.address = next_address;
		info.name = format("%08X", bits);
		info.id = bits;
		queue.push(info);

		cmd.command[0] = MFX_ASSIGN_ADDRESS;
		cmd.command[1] = next_address>>8;
		cmd.command[2] = next_address;
		for(unsigned i=0; i<4; ++i)
			cmd.command[3+i] = bits>>(24-i*8);
		cmd.length = 7;

		cmd.tag.type = Tag::GENERAL;
		cmd.tag.command = NEW_LOCO;
		cmd.tag.id = bits;

		size = 0;
		bits = 0;
		++next_address;

		return true;
	}

	cmd.command[0] = MFX_SEARCH;
	for(unsigned i=0; i<4; ++i)
		cmd.command[1+i] = bits>>(24-i*8);
	cmd.command[5] = size;
	cmd.length = 6;

	sleep(200*Time::msec);

	if(control.debug>=1)
		IO::print("Search %08X/%d\n", bits, size);

	return true;
}

void ArduControl::MfxSearchTask::process_reply(const char *reply, unsigned length)
{
	unsigned char type = reply[0];
	if(type==MFX_SEARCH_FEEDBACK && length==2)
	{
		if(reply[1])
		{
			misses = 0;
			++size;
		}
		else if(size>0 && misses<6)
		{
			++misses;
			bits ^= 1<<(32-size);
		}
		else
		{
			sleep(2*Time::sec);
			bits = 0;
			size = 0;
			misses = 0;
		}
	}
}

void ArduControl::MfxSearchTask::set_next_address(unsigned a)
{
	next_address = a;
}

bool ArduControl::MfxSearchTask::pop_info(MfxInfo &info)
{
	return queue.pop(info);
}


ArduControl::MonitorTask::MonitorTask():
	Task("Monitor"),
	voltage(0),
	current(0),
	base_level(0),
	peak_level(0),
	next_type(0)
{ }

bool ArduControl::MonitorTask::get_work(PendingCommand &cmd)
{
	if(next_type==0)
		cmd.command[0] = READ_INPUT_VOLTAGE;
	else
		cmd.command[0] = READ_TRACK_CURRENT;
	cmd.length = 1;

	sleep(200*Time::msec);
	next_type = (next_type+1)%5;

	return true;
}

void ArduControl::MonitorTask::process_reply(const char *reply, unsigned length)
{
	unsigned char type = reply[0];
	if(type==INPUT_VOLTAGE && length==3)
		voltage = ((static_cast<unsigned char>(reply[1])<<8) | static_cast<unsigned char>(reply[2]))/1000.0f;
	else if(type==TRACK_CURRENT && length==5)
	{
		current = ((static_cast<unsigned char>(reply[1])<<8) | static_cast<unsigned char>(reply[2]))/1000.0f;
		float peak = ((static_cast<unsigned char>(reply[3])<<8) | static_cast<unsigned char>(reply[4]))/1000.0f;
		peak_level = max(peak_level, peak);
		base_level = min(base_level, current);
	}
}

void ArduControl::MonitorTask::reset_peak()
{
	base_level = current;
	peak_level = current;
}


ArduControl::ControlThread::ControlThread(ArduControl &c):
	control(c),
	done(false)
{
	tasks.push_back(&control.command_queue);
	tasks.push_back(&control.monitor);
	tasks.push_back(&control.mfx_announce);
	tasks.push_back(&control.mfx_search);
	tasks.push_back(&control.s88);
	tasks.push_back(&control.refresh);

	launch();
}

void ArduControl::ControlThread::exit()
{
	done = true;
	join();
}

void ArduControl::ControlThread::main()
{
	init_baud_rate();

	while(!done)
	{
		PendingCommand cmd;
		if(get_work(cmd))
		{
			bool success = true;
			bool resync = false;
			for(unsigned i=0; (success && i<cmd.repeat_count); ++i)
			{
				unsigned result = do_command(cmd, control.command_timeout);
				success = (result==COMMAND_OK);
				resync = (result==0);
			}

			if(success && cmd.tag)
				control.completed_commands.push(cmd.tag);

			if(resync)
			{
				if(control.debug>=1)
					IO::print("Synchronization with ArduControl lost, attempting to recover\n");
				for(unsigned i=0; (resync && i<16); ++i)
				{
					control.serial.put('\xFF');
					while(IO::poll(control.serial, IO::P_INPUT, control.command_timeout))
						resync = (control.serial.get()!=0xFF);
				}
				if(resync)
				{
					if(control.debug>=1)
						IO::print("Resynchronization failed, giving up\n");
					done = true;
				}
				else
				{
					if(control.debug>=1)
						IO::print("Resynchronization successful\n");
					if(cmd.tag)
						control.command_queue.push(cmd);
				}
			}
		}
		else
			Time::sleep(10*Time::msec);
	}
}

void ArduControl::ControlThread::init_baud_rate()
{
	static unsigned rates[] = { 57600, 9600, 19200, 38400, 0 };
	unsigned rate = 0;
	control.serial.set_data_bits(8);
	control.serial.set_parity(IO::Serial::NONE);
	control.serial.set_stop_bits(1);
	for(unsigned i=0; rates[i]; ++i)
	{
		control.serial.set_baud_rate(rates[i]);
		control.serial.put('\xFF');
		if(IO::poll(control.serial, IO::P_INPUT, 500*Time::msec))
		{
			int c = control.serial.get();
			if(c==0xFF)
			{
				rate = rates[i];
				break;
			}
		}
	}

	if(!rate)
	{
		if(control.debug>=1)
			IO::print("ArduControl detection failed\n");
		done = true;
		return;
	}

	if(control.debug>=1)
		IO::print("ArduControl detected at %d bits/s\n", rate);

	if(rate!=rates[0])
	{
		PendingCommand cmd;
		cmd.command[0] = SET_BAUD_RATE;
		cmd.command[1] = rates[0]>>8;
		cmd.command[2] = rates[0];
		cmd.length = 3;
		if(do_command(cmd, Time::sec)==COMMAND_OK)
		{
			control.serial.set_baud_rate(rates[0]);
			Time::sleep(Time::sec);
			if(do_command(cmd, Time::sec)==COMMAND_OK)
			{
				if(control.debug>=1)
					IO::print("Rate changed to %d bits/s\n", rates[0]);
			}
		}
	}
}

bool ArduControl::ControlThread::get_work(PendingCommand &cmd)
{
	Time::TimeStamp t = Time::now();

	unsigned count = 0;
	for(; (count<tasks.size() && tasks[count]->get_sleep_timeout()<=t); ++count) ;

	for(; count>0; --count)
	{
		unsigned i = 0;
		for(unsigned j=1; j<count; ++j)
			if(tasks[j]->get_priority()<tasks[i]->get_priority())
				i = j;

		Task *task = tasks[i];
		bool result = task->get_work(cmd);

		Time::TimeStamp st = max(task->get_sleep_timeout(), t);
		for(; (i+1<tasks.size() && tasks[i+1]->get_sleep_timeout()<=st); ++i)
			tasks[i] = tasks[i+1];
		tasks[i] = task;

		if(result)
		{
			if(control.debug>=2)
				IO::print("Scheduled task %s\n", task->get_name());
			return true;
		}
	}

	// As fallback, send an idle packet for the MM protocol
	cmd.command[0] = MOTOROLA_SPEED;
	cmd.command[1] = 80;
	cmd.command[2] = 0;
	cmd.command[3] = 0;
	cmd.length = 4;

	return true;
}

unsigned ArduControl::ControlThread::do_command(const PendingCommand &cmd, const Time::TimeDelta &timeout)
{
	if(control.debug>=2)
	{
		string cmd_hex;
		for(unsigned i=0; i<cmd.length; ++i)
			cmd_hex += format(" %02X", static_cast<unsigned char>(cmd.command[i]));
		IO::print("< %02X%s\n", cmd.length^0xFF, cmd_hex);
	}

	control.serial.put(cmd.length^0xFF);
	control.serial.write(cmd.command, cmd.length);

	unsigned result = 0;
	while(1)
	{
		bool got_data;
		if(result)
			got_data = IO::poll(control.serial, IO::P_INPUT, Time::zero);
		else
			got_data = IO::poll(control.serial, IO::P_INPUT, timeout);

		if(!got_data)
			break;

		unsigned rlength = control.serial.get()^0xFF;
		if(rlength>15)
		{
			IO::print("Invalid length %02X\n", rlength);
			continue;
		}

		char reply[15];
		unsigned pos = 0;
		while(pos<rlength)
		{
			if(!IO::poll(control.serial, IO::P_INPUT, timeout))
				return 0;
			pos += control.serial.read(reply+pos, rlength-pos);
		}

		if(control.debug>=2)
		{
			string reply_hex;
			for(unsigned i=0; i<rlength; ++i)
				reply_hex += format(" %02X", static_cast<unsigned char>(reply[i]));
			IO::print("> %02X%s\n", rlength^0xFF, reply_hex);
		}

		unsigned r = process_reply(reply, rlength);
		if(r && !result)
			result = r;
	}

	return result;
}

unsigned ArduControl::ControlThread::process_reply(const char *reply, unsigned rlength)
{
	unsigned char type = reply[0];
	if((type&0xE0)==0x80)
	{
		if(type!=COMMAND_OK)
			IO::print("Error %02X\n", type);
		return type;
	}
	else if(type==POWER_STATE && rlength==2)
		set_power(reply[1]);
	else if(type==OVERCURRENT)
	{
		set_power(false);
		IO::print("Overcurrent detected!\n");
	}
	else
	{
		for(vector<Task *>::iterator i=tasks.begin(); i!=tasks.end(); ++i)
			(*i)->process_reply(reply, rlength);
	}

	return 0;
}

void ArduControl::ControlThread::set_power(bool p)
{
	control.power.set(p);

	Tag tag;
	tag.type = Tag::GENERAL;
	tag.command = POWER;
	tag.serial = control.power.serial;
	control.completed_commands.push(tag);
}


ArduControl::Loader::Loader(ArduControl &c):
	DataFile::ObjectLoader<ArduControl>(c)
{
	add("mfx_announce_serial", &Loader::mfx_announce_serial);
	add("mfx_locomotive", &Loader::mfx_locomotive);
}

void ArduControl::Loader::mfx_announce_serial(unsigned s)
{
	obj.mfx_announce.set_serial(s);
}

void ArduControl::Loader::mfx_locomotive(unsigned id)
{
	MfxInfo info;
	info.id = id;
	info.protocol = "MFX";
	load_sub(info);
	obj.add_mfx_info(info);
}


ArduControl::MfxInfo::Loader::Loader(MfxInfo &i):
	DataFile::ObjectLoader<MfxInfo>(i)
{
	add("address", static_cast<unsigned MfxInfo::*>(&MfxInfo::address));
	add("name", static_cast<string MfxInfo::*>(&MfxInfo::name));
}

} // namespace R2C2
