#include <msp/datafile/directorysource.h>
#include <msp/datafile/packsource.h>
#include <msp/datafile/parser.h>
#include <msp/fs/dir.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include "catalogue.h"
#include "signaltype.h"
#include "terraintype.h"
#include "trackappearance.h"
#include "tracktype.h"
#include "vehicletype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

Catalogue::Catalogue():
	scale(1),
	layout(*this)
{
	add_type<TrackType>().keyword("track");
	add_type<SignalType>().keyword("signal");
	add_type<VehicleType>().suffix(".veh").keyword("vehicle");
	add_type<TerrainType>().suffix(".terr").keyword("terrain");
	add_type<TrackAppearance>().suffix("trkapp").keyword("track_appearance");

	add_source("data/r2c2");
}

void Catalogue::add_source(const FS::Path &path)
{
	string ext = FS::extpart(path.str());
	DataFile::CollectionSource *source = 0;
	IO::Seekable *init_file = 0;
	if(ext==".mdt")
	{
		DataFile::PackSource *pack_source = new DataFile::PackSource;
		pack_source->add_pack_file(path.str());
		source = pack_source;

		list<DataFile::PackSource::FileInfo> files = pack_source->list_files();
		for(list<DataFile::PackSource::FileInfo>::const_iterator i=files.begin(); i!=files.end(); ++i)
		{
			string fext = FS::extpart(i->name);
			if(fext==".cat" && !init_file)
				init_file = pack_source->open(i->name);
		}
	}
	else if(FS::is_dir(path))
	{
		DataFile::DirectorySource *dir_source = new DataFile::DirectorySource;
		dir_source->add_directory(path);

		list<FS::Path> queue;
		queue.push_back(path);
		while(!queue.empty())
		{
			FS::Path dir = queue.front();
			queue.pop_front();
			list<string> files = FS::list_files(dir);
			for(list<string>::const_iterator i=files.begin(); i!=files.end(); ++i)
			{
				FS::Path full = dir/ *i;
				if(FS::is_dir(full))
				{
					dir_source->add_directory(full);
					queue.push_back(full);
				}

				string fext = FS::extpart(*i);
				if(fext==".cat" && !init_file)
					init_file = new IO::BufferedFile(full.str());
			}
		}

		source = dir_source;
	}

	sources.push_back(source);
	Collection::add_source(*source);
	signal_source_added.emit(*source);

	if(init_file)
	{
		DataFile::Parser parser(*init_file, path.str());
		Loader ldr(*this);
		ldr.load(parser);
	}
	delete init_file;
}


Catalogue::Loader::Loader(Catalogue &c):
	Collection::Loader(c),
	cat(c)
{
	add("layout", &Loader::layout);
	add("scale", &Loader::scale);
}

void Catalogue::Loader::layout()
{
	load_sub(cat.layout);
}

void Catalogue::Loader::scale(float n, float d)
{
	cat.scale = n/d;
}

} // namespace R2C2
