#include <cmath>
#include <msp/datafile/collection.h>
#include <msp/geometry/union.h>
#include "trackappearance.h"
#include "tracktype.h"

using namespace std;
using namespace Msp;

namespace R2C2 {

TrackType::TrackType():
	appearance(0),
	state_bits(0),
	autofit_preference(1)
{ }

const TrackAppearance &TrackType::get_appearance() const
{
	if(!appearance)
		throw logic_error("no appearance");
	return *appearance;
}

float TrackType::get_gauge() const
{
	return get_appearance().get_gauge();
}

float TrackType::get_total_length() const
{
	return get_path_length(-1);
}

float TrackType::get_path_length(int p) const
{
	float len = 0;
	for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
		if(p<0 || i->get_path()==static_cast<unsigned>(p))
			len += i->get_length();
	return len;
}

unsigned TrackType::get_paths() const
{
	unsigned mask = 0;
	for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
		mask |= 1<<i->get_path();
	return mask;
}

unsigned TrackType::get_n_paths() const
{
	unsigned n = 0;
	for(unsigned mask = get_paths(); mask; ++n)
		mask &= mask-1;
	return n;
}

unsigned TrackType::coerce_path(unsigned entry, unsigned path) const
{
	const Endpoint &ep = get_endpoint(entry);
	if(ep.has_path(path))
		return path;

	unsigned paths = get_paths();
	if(paths>>(1<<state_bits))
	{
		/* There are more paths than can be expressed with state_bits, so
		multiple paths are set at once.  See if one of the alternatives fits. */
		unsigned step = 1<<state_bits;
		for(unsigned p=path+step; paths>>p; p+=step)
			if(ep.has_path(p))
				return p;
	}

	// Find an endpoint that's connected to the entry and has the requested path
	for(vector<Endpoint>::const_iterator i=endpoints.begin(); i!=endpoints.end(); ++i)
		if(i->has_path(path) && i->has_common_paths(ep))
		{
			unsigned p = 1;
			for(unsigned m=i->paths&ep.paths; m>>p; ++p) ;
			return p-1;
		}

	// TODO crossings fall here
	throw logic_error("TrackType::coerce_path");
}

bool TrackType::is_turnout() const
{
	return endpoints.size()>2;
}

bool TrackType::is_dead_end() const
{
	return endpoints.size()<2;
}

const TrackType::Endpoint &TrackType::get_endpoint(unsigned i) const
{
	if(i>=endpoints.size())
		throw out_of_range("TrackType::get_endpoint");

	return endpoints[i];
}

OrientedPoint TrackType::get_point(unsigned epi, unsigned path, float d) const
{
	if(epi>=endpoints.size())
		throw out_of_range("TrackType::get_point");

	const TrackPart *part = 0;
	unsigned part_ep = 0;
	for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
	{
		if(endpoints[epi].has_path(path) && i->get_path()!=path)
			continue;

		unsigned n_part_eps = (i->is_dead_end() ? 1 : 2);
		for(unsigned j=0; j<n_part_eps; ++j)
		{
			OrientedPoint p = i->get_point(j ? i->get_length() : 0);
			Vector span = p.position-endpoints[epi].pos;
			if(dot(span, span)<1e-6)
			{
				part = &*i;
				part_ep = j;
			}
		}
	}

	if(!part)
		throw logic_error("internal error (endpoint does not match any part)");

	while(1)
	{
		float plen = part->get_length();
		if(d<=plen)
		{
			if(part_ep==1)
				d = plen-d;
			OrientedPoint p = part->get_point(d);
			if(part_ep==1)
				p.rotation += Angle::half_turn();
			return p;
		}
		else
		{
			d -= plen;
			TrackPart *next = part->get_link(1-part_ep);
			if(!next)
				throw invalid_argument("TrackType::get_point");
			part_ep = (next->get_link(0)==part ? 0 : 1);
			part = next;
		}
	}
}

OrientedPoint TrackType::get_nearest_point(const Vector &p) const
{
	OrientedPoint result;
	float dist = -1;

	for(vector<TrackPart>::const_iterator i=parts.begin(); i!=parts.end(); ++i)
	{
		OrientedPoint n = i->get_nearest_point(p);
		float d = distance(n.position, p);
		if(d<dist || dist<0)
		{
			result = n;
			dist = d;
		}
	}

	return result;
}

void TrackType::collect_endpoints()
{
	endpoints.clear();

	for(vector<TrackPart>::iterator i=parts.begin(); i!=parts.end(); ++i)
	{
		for(vector<TrackPart>::iterator j=i; ++j!=parts.end();)
			i->check_link(*j);

		unsigned n_part_eps = (i->is_dead_end() ? 1 : 2);
		for(unsigned j=0; j<n_part_eps; ++j)
			if(!i->get_link(j))
			{
				OrientedPoint p = i->get_point(j ? i->get_length() : 0);
				if(j==0)
					p.rotation += Angle::half_turn();

				bool found = false;
				for(vector<Endpoint>::iterator k=endpoints.begin(); k!=endpoints.end(); ++k)
				{
					Vector d = k->pos-p.position;

					Angle da = wrap_balanced(k->dir-p.rotation);

					if(dot(d, d)<1e-6 && abs(da).radians()<0.01)
					{
						k->paths |= 1<<i->get_path();
						found = true;
						break;
					}
				}

				if(!found)
					endpoints.push_back(Endpoint(p.position.x, p.position.y, p.rotation, 1<<i->get_path()));
			}
	}
}

TrackType::Endpoint::Endpoint(float x, float y, const Angle &d, unsigned p):
	pos(x, y, 0),
	dir(d),
	paths(p)
{ }


TrackType::Loader::Loader(TrackType &t, Collection &c):
	DataFile::DerivedObjectLoader<TrackType, ObjectType::Loader>(t),
	coll(c),
	state_bits_set(false)
{
	add("appearance",  &TrackType::appearance);
	add("autofit_preference", &TrackType::autofit_preference);
	add("object",      &TrackType::object);
	add("state_bits",  &Loader::state_bits);
	add("part",        &Loader::part);
}

void TrackType::Loader::finish()
{
	obj.collect_endpoints();
	vector<const Shape *> shapes;
	for(vector<TrackPart>::iterator i=obj.parts.begin(); i!=obj.parts.end(); ++i)
	{
		i->create_shape();
		shapes.push_back(&i->get_shape());
	}
	obj.shape = Geometry::Union<float, 3>::from_iterator_range(shapes.begin(), shapes.end()).clone();
}

void TrackType::Loader::part()
{
	TrackPart p;
	load_sub(p);
	obj.parts.push_back(p);
	if(!state_bits_set && p.get_path())
		while(p.get_path()>=(1U<<obj.state_bits))
			++obj.state_bits;
}

void TrackType::Loader::state_bits(unsigned b)
{
	obj.state_bits = b;
	state_bits_set = true;
}

} // namespace R2C2
